﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Browser;
using System.Xml;

namespace DeepZoomTravelDemo
{
    public partial class Page : UserControl
    {
        private bool _Dragging = false;
        private Point _LastViewportOrigin;
        private Point _LastMousePosition;
        private SubImageInfo[] _Metadata;
        private int _LastIndex = -1;

        public Page()
        {
            InitializeComponent();

            // Register mousewheel event handler
            HtmlPage.Window.AttachEvent("DOMMouseScroll", OnMouseWheelTurned);
            HtmlPage.Window.AttachEvent("onmousewheel", OnMouseWheelTurned);
            HtmlPage.Document.AttachEvent("onmousewheel", OnMouseWheelTurned);

            // Fetch Metadata.xml from the server
            WebClient wc = new WebClient();
            wc.OpenReadCompleted += new OpenReadCompletedEventHandler(WebClient_OpenReadCompleted);
            wc.OpenReadAsync(new Uri("Metadata.xml", UriKind.Relative));
        }

        private void WebClient_OpenReadCompleted(object sender, OpenReadCompletedEventArgs e)
        {
            if (e.Error != null)
            {
                MessageBox.Show("Unable to load XML metadata");
                return;
            }
        
            // Create a collection of SubImageInfo objects from Metadata.xml
            List<SubImageInfo> images = new List<SubImageInfo>();

            try
            {
                XmlReader reader = XmlReader.Create(e.Result);
                SubImageInfo info = null;

                while (reader.Read())
                {
                    if (reader.NodeType == XmlNodeType.Element && reader.Name == "Image")
                        info = new SubImageInfo();
                    else if (reader.NodeType == XmlNodeType.Element && reader.Name == "ZOrder")
                        info.Index = reader.ReadElementContentAsInt();
                    else if (reader.NodeType == XmlNodeType.Element && reader.Name == "Tag")
                    {
                        string[] substrings = reader.ReadElementContentAsString().Split('+');
                        info.Caption = substrings[0];
                        if (substrings.Length > 1)
                            info.Description = substrings[1];
                        else
                            info.Description = String.Empty;
                    }
                    else if (reader.NodeType == XmlNodeType.EndElement && reader.Name == "Image")
                        images.Add(info);
                }
            }
            catch (XmlException)
            {
                MessageBox.Show("Error parsing XML metadata");
            }

            // Populate the _Metadata array with ordered data
            _Metadata = new SubImageInfo[images.Count];

            foreach (SubImageInfo image in images)
                _Metadata[image.Index - 1] = image;
        }

        private void MSI_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            _LastViewportOrigin = MSI.ViewportOrigin;
            _LastMousePosition = e.GetPosition(MSI);
            _Dragging = true;
        }

        private void MSI_MouseMove(object sender, MouseEventArgs e)
        {
            if (_Dragging)
            {
                // If the left mouse button is down, pan the Deep Zoom scene
                Point pos = e.GetPosition(MSI);
                Point origin = _LastViewportOrigin;
                origin.X += (_LastMousePosition.X - pos.X) / MSI.ActualWidth * MSI.ViewportWidth;
                origin.Y += (_LastMousePosition.Y - pos.Y) / MSI.ActualWidth * MSI.ViewportWidth;
                MSI.ViewportOrigin = _LastViewportOrigin = origin;
                _LastMousePosition = pos;
            }
            else
            {
                // If the left mouse button isn't down, update the infobar
                if (_Metadata != null)
                {
                    int index = GetSubImageIndex(e.GetPosition(MSI));

                    if (index != _LastIndex)
                    {
                        _LastIndex = index;

                        if (index != -1)
                        {
                            Caption.Text = _Metadata[index].Caption;
                            Description.Text = _Metadata[index].Description;
                            FadeIn.Begin();
                        }
                        else
                        {
                            FadeOut.Begin();
                        }
                    }
                }
            }
        }

        private void MSI_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            _Dragging = false;
        }

        private void MSI_MouseLeave(object sender, MouseEventArgs e)
        {
            _Dragging = false;
        }

        private void OnMouseWheelTurned(Object sender, HtmlEventArgs args)
        {
            double delta = 0;
            ScriptObject e = args.EventObject;

            if (e.GetProperty("wheelDelta") != null) // IE and Opera
            {
                delta = ((double)e.GetProperty("wheelDelta"));
                if (HtmlPage.Window.GetProperty("opera") != null)
                    delta = -delta;
            }
            else if (e.GetProperty("detail") != null) // Mozilla and Safari
            {
                delta = -((double)e.GetProperty("detail"));
            }

            if (delta > 0)
            {
                ZoomIn(args.ClientX, args.ClientY);
            }
            else if (delta < 0)
            {
                ZoomOut(args.ClientX, args.ClientY);
            }

            if (delta != 0)
            {
                args.PreventDefault();
                e.SetProperty("returnValue", false);
            }
        }

        private void ZoomIn(double x, double y)
        {
            if (!_Dragging)
            {
                Point point = MSI.ElementToLogicalPoint(new Point(x, y));
                MSI.ZoomAboutLogicalPoint(1.25, point.X, point.Y);
            }
        }

        private void ZoomOut(double x, double y)
        {
            if (!_Dragging)
            {
                Point point = MSI.ElementToLogicalPoint(new Point(x, y));
                MSI.ZoomAboutLogicalPoint(0.8, point.X, point.Y);
            }
        }

        private void TextBlock_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            Shuffle();
        }

        private int GetSubImageIndex(Point point)
        {
            // Hit-test each subimage in the MultiScaleImage control to determine
            // whether "point" lies within a subimage
            for (int i = MSI.SubImages.Count -1; i >= 0; i--)
            {
                MultiScaleSubImage image = MSI.SubImages[i];
                double width = MSI.ActualWidth / (MSI.ViewportWidth * image.ViewportWidth);
                double height = MSI.ActualWidth / (MSI.ViewportWidth * image.ViewportWidth * image.AspectRatio);

                Point pos = MSI.LogicalToElementPoint(new Point(
                    -image.ViewportOrigin.X / image.ViewportWidth,
                    -image.ViewportOrigin.Y / image.ViewportWidth)
                );
                Rect rect = new Rect(pos.X, pos.Y, width, height);

                if (rect.Contains(point))
                {
                    // Return the image index
                    return i;
                }
            }

            // No corresponding subimage
            return -1;
        }

        private void Shuffle()
        {
            // Create a randomly ordered list of subimage viewport origins
            List<Point> origins = new List<Point>();

            foreach (MultiScaleSubImage image in MSI.SubImages)
                origins.Add(image.ViewportOrigin);

            Random rand = new Random();
            int count = origins.Count;

            for (int i = 0; i < count; i++)
            {
                Point origin = origins[i];
                origins.RemoveAt(i);
                origins.Insert(rand.Next(count), origin);
            }

            // Create a Storyboard and animations for shuffling
            Storyboard sb = new Storyboard();

            for (int i = 0; i < count; i++)
            {
                PointAnimation animation = new PointAnimation();
                animation.Duration = TimeSpan.FromMilliseconds(250);
                animation.To = origins[i];
                Storyboard.SetTarget(animation, MSI.SubImages[i]);
                Storyboard.SetTargetProperty(animation, new PropertyPath("ViewportOrigin"));
                sb.Children.Add(animation);
            }

            // Run the animations
            sb.Begin();        
        }
    }
}
